//provides tcp remote-logging capability

#include "console-server.h"
#include "common.h"

pthread_mutex_t console_server::console_server_lock = PTHREAD_MUTEX_INITIALIZER;
int console_server::listener_socket = 0;
	
/*static */ void *console_server::listener_thread( void *arg )
{
	int i, fd;
	socklen_t remote_endpoint_length = sizeof(struct sockaddr_in);
	struct sockaddr_in listener_endpoint, remote_endpoint;

	console_server *server = (console_server *)arg;
	
	// setup cancellation behavior
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

	if((console_server::listener_socket = socket(PF_INET, SOCK_STREAM, 0)) < 0)
	{
		_slog(NO_TRANS, _DEBUG_BASIC, 1, "[Console Server]: Error obtaining listener socket.", 2);
		return NULL;
	}
	
	listener_endpoint.sin_family = AF_INET;
	listener_endpoint.sin_port = _htons(server->port);
	listener_endpoint.sin_addr.s_addr = _htonl(INADDR_ANY);

	// -- setting this option allows for the listening socket to be immediately available after it's
	// -- been close (i.e., when rqcamd has been restarted, otherwise listening socket binding
	// -- errors may appear at restart.
	int reuse = 1;
	if( setsockopt(console_server::listener_socket, SOL_SOCKET, SO_REUSEADDR, &reuse, sizeof(reuse) ) < 0 )
	{
		_slog(NO_TRANS, _DEBUG_BASIC, 1, "[Console Server]: Error configuring listening socket.", 2);
		return NULL;
	}

	if( bind(console_server::listener_socket, (struct sockaddr *)&listener_endpoint, sizeof(listener_endpoint)) < 0 )
	{
		_slog(NO_TRANS, _DEBUG_BASIC, 1, "[Console Server]: Error binding listening socket to local interface.", 2);
		return NULL;
	}
	
	if(listen(console_server::listener_socket, _MAX_CONCURRENT_CONNECTIONS) < 0)
	{
		_slog(NO_TRANS, _DEBUG_BASIC, 1, "[Console Server]: Error listening.", 2);
		return NULL;
	}
	
	//listening loop
	while( 1 )
	{
		fd = accept(console_server::listener_socket, (struct sockaddr *)&remote_endpoint, &remote_endpoint_length);
	
		//lock server, assign fd to connection array
		pthread_mutex_lock( &console_server::console_server_lock );
		for(i=0;i<_MAX_CONCURRENT_CONNECTIONS;i++)
		{
			if( server->fds[i] < 0 )
			{
				server->fds[i] = fd;
				break;
			}
		}
		pthread_mutex_unlock( &console_server::console_server_lock );
	
		if( i == _MAX_CONCURRENT_CONNECTIONS )
		{
			char message[] = "[Console Server] Maximum number of concurrent connections reached.";
			network_write(fd, message, strlen(message));
			close( fd );
		}
	}
}

/* static */ void console_server::listener_thread_cleanup( void )
{
	close(console_server::listener_socket);
}

console_server::console_server( unsigned int listening_port )
{
	listener = 0;
	port = listening_port;
	
	if( pthread_create(&listener, NULL, console_server::listener_thread, this) )
	{
		_slog(NO_TRANS, _DEBUG_BASIC, 1, "[Console Server]: Error starting console broardcast listener thread.", 2);
	}
	
	for(int i=0;i<_MAX_CONCURRENT_CONNECTIONS;i++) fds[i] = -1;
}

console_server::~console_server()
{
	//stop listener thread
	pthread_cancel(listener);
	pthread_join(listener, NULL);
	console_server::listener_thread_cleanup();

	//shutdown all current connections
	pthread_mutex_lock( &console_server::console_server_lock );
	for(int i=0;i<_MAX_CONCURRENT_CONNECTIONS;i++)
	{
		if( fds[i] >= 0 )
		{
			close( fds[i] );
			fds[i] = -1;
		}
	}
	pthread_mutex_unlock( &console_server::console_server_lock );
}

void console_server::broadcast(char *message)
{
	//converted_message[0] = 0;
	//broadcast message to all active connections. on failed writes, close connections.
	pthread_mutex_lock( &console_server::console_server_lock );
	for(int i=0;i<_MAX_CONCURRENT_CONNECTIONS;i++)
	{
		if( fds[i] >= 0 )
		{
			unix2dos(converted_message, message);
			if( network_write(fds[i], converted_message, strlen(converted_message)) < 0 )
			{
				close( fds[i] );
				fds[i] = -1;
			}
		}
	}
	pthread_mutex_unlock( &console_server::console_server_lock );
}

/* static */ int console_server::network_write(int fd, char *buffer, unsigned int length)
{
	//write with 1 sec time-out to avoid emu lock-ups due to unreliable client connections.
	struct pollfd pfd;
		
	pfd.fd = fd;
	pfd.events = POLLOUT | POLLERR;
	pfd.revents = 0;
	
	if(poll(&pfd, 1, 2000) != 1) return -1;
	
	if(pfd.revents & POLLOUT)
		return write(fd, buffer, length);
	else
		return -1;
}
